@@ -1,32 +1,25 @@
-# requests is an optional library that can be found at http://docs.python-requests.org/en/latest/
 """This Python module provides some basic services for interacting with ScienceBase."""
 from __future__ import print_function
-try:
-    # For Python 3.0 and later
-    import http.client as httplib
-    from urllib.request import urlopen
-    from urllib.parse import urlencode
-    import urllib.parse as urlparse
 
-except ImportError:
-    # Fall back to Python 2's urllib2
-    import httplib
-    from urllib import urlencode
-    from urllib2 import urlopen
-    import urlparse
-
-import requests
+# For Python 3.0 and later
+import http.client as httplib
+from urllib.parse import urlencode
+import urllib.parse as urlparse
+import sys
+import logging
 import json
 import os
 import getpass
-import logging
 import mimetypes
+import requests
+
 from pkg_resources import get_distribution
 from pkg_resources import DistributionNotFound
+from sb3.SbSessionEx import SbSessionEx
 
 class SbSession:
-    """SbSession encapsulates a session with ScienceBase, and provides methods for working with ScienceBase Catalog
-    Items.
+    """SbSession encapsulates a session with ScienceBase, and provides methods for working with
+    ScienceBase Catalog Items.
     """
     _josso_url = None
     _base_sb_url = None
@@ -49,7 +42,8 @@ class SbSession:
     _retry = False
     _max_item_count = 1000
     _env = None
-    
+    _sbSessionEx = None
+
     def __init__(self, env=None):
         """Initialize session and set JSON headers"""
         self._env = env
@@ -78,13 +72,13 @@ class SbSession:
         self._base_undelete_item_url = self._base_item_url + "undelete/"
         self._base_shortcut_item_url = self._base_items_url + "addLink/"
         self._base_unlink_item_url = self._base_items_url + "unlink/"
-        self._base_person_url = self._base_directory_url + "person/"        
+        self._base_person_url = self._base_directory_url + "person/"
 
         self._session = requests.Session()
         self._session.headers.update({'Accept': 'application/json'})
         sciencebasepy_agent = ' sciencebase-sciencebasepy'
         try:
-            sciencebasepy_agent += '/%s' % get_distribution("sciencebasepy").version
+            sciencebasepy_agent += f'/{get_distribution("sciencebasepy").version}'
         except DistributionNotFound:
             pass
         self._session.headers.update({'User-Agent': self._session.headers['User-Agent'] + sciencebasepy_agent})
@@ -114,6 +108,9 @@ class SbSession:
         self._jossosessionid = self._session.cookies['JOSSO_SESSIONID']
         self._session.headers.update({'MYUSGS-JOSSO-SESSION-ID': self._jossosessionid})
 
+        # Login to Keycloak for SB3 calls
+        self._sbSessionEx = SbSessionEx(self._env).login(username, password)
+
         return self
 
     def logout(self):
@@ -122,21 +119,20 @@ class SbSession:
         self._session.cookies.clear_session_cookies()
         self._session.params = {}
 
-    def loginc(self, username):
+    def loginc(self, username, tries=3):
         """Log into ScienceBase, prompting for the password
         
         :param username: The ScienceBase user to log in as
         :return: The SbSession object with the user logged in
         """
-        tries = 0
-        while (tries < 5):
+        while tries > 0:
             password = getpass.getpass()
             try:
                 return self.login(username, password)
             except Exception:
-                tries += 1
+                tries -= 1
                 print("Invalid password, try again")
-        raise Exception("Too many invalid password attemps, you may need to wait 15 minutes before trying again")
+        raise Exception("Too many invalid password attemps")
 
     def is_logged_in(self):
         """Determine whether the SbSession is logged in and active in ScienceBase
@@ -294,15 +290,6 @@ class SbSession:
         ret = self._session.put(self._base_item_url + item_json['id'], data=json.dumps(item_json))
         return self._get_json(ret)
 
-    def update_items(self, items_json):
-        """Update existing ScienceBase Items
-
-        :param item_json: JSON list representing the ScienceBase Catalog items to update
-        :return: Full item JSON from ScienceBase Catalog after update
-        """
-        ret = self._session.put(self._base_items_url + "upsert/", data=json.dumps(items_json))
-        return self._get_json(ret)
-
     def update_hidden_property(self, item_id, hidden_property_id, hidden_property_json):
         """Update an existing hidden property of a ScienceBase Item
 
@@ -397,13 +384,32 @@ class SbSession:
     def upload_file_to_item(self, item, filename, scrape_file=True):
         """Upload a file to an existing Item in ScienceBase
 
-        :param item:
+        :param item: ScienceBase Catalog Item JSON of the Item to update
         :param filename: Filenames of the file to upload
         :param scrape_file: Whether to scrape metadata and create extensions from special files
         :return: The ScienceBase Catalog Item JSON of the updated Item
         """
         return self.upload_files_and_update_item(item, [filename], scrape_file)
 
+    def upload_cloud_file_to_item(self, itemid, filename):
+        """Upload a file to cloud storage on an existing Item in ScienceBase
+
+        :param itemid: ScienceBase Catalog Item ID of the Item to update
+        :param filename: Filenames of the file to upload
+        :return: The ScienceBase Catalog Item JSON of the updated Item
+        """
+        ret = None
+        if not self._sbSessionEx.is_logged_in():
+            print(f'{self._username} not logged into Keycloak -- cloud services not available')
+        else:
+            mimetype = self._guess_mimetype(filename)
+            response = self._sbSessionEx.upload_cloud_file_upload_session(itemid, filename, mimetype)
+            if 'data' in response and 'completeMultiPartUpload' in response['data'] and 'Successful' in response['data']['completeMultiPartUpload']:
+                ret = self.get_item(itemid)
+            else:
+                raise Exception('Cloud upload failed for', filename)
+        return ret
+
     def upload_file_and_create_item(self, parentid, filename, scrape_file=True):
         """Upload a file and create a new Item in ScienceBase
 
@@ -412,7 +418,7 @@ class SbSession:
         :param scrape_file: Whether to scrape metadata and create extensions from special files
         :return: The ScienceBase Catalog Item JSON of the new Item
         """
-        return self.upload_files_and_create_item(parentid, [filename], scrape_file)
+        return self.upload_files_and_create_item(parentid, [filename], scrape_file)    
 
     def upload_files_and_create_item(self, parentid, filenames, scrape_file=True):
         """Upload multiple files and create a new Item in ScienceBase
@@ -445,7 +451,7 @@ class SbSession:
         url = self._base_upload_file_url
         files = []
         for filename in filenames:
-            if type(filename) == str:
+            if isinstance(filename, str):
                 if (os.access(filename, os.F_OK)):
                     files.append(('file', open(filename, 'rb')))
                 else:
@@ -471,15 +477,16 @@ class SbSession:
         retval = None
         url = self._base_upload_file_temp_url
 
-        if (os.access(filename, os.F_OK)):
+        if os.access(filename, os.F_OK):
             #
             # if no mimetype was sent in, try to guess
             #
-            if None == mimetype:
-                mimetype = mimetypes.guess_type(filename)
-            (path, fname) = os.path.split(filename)
-            ret = self._session.post(url, files=[('files[]', (fname, open(filename, 'rb'), mimetype))])
-            retval = self._get_json(ret)
+            if mimetype is None:
+                mimetype = self._guess_mimetype(filename)
+            fname = os.path.basename(filename)
+            with open(filename, 'rb') as f:
+                ret = self._session.post(url, files=[('files[]', (fname, f, mimetype))])
+                retval = self._get_json(ret)
         else:
             raise Exception("File not found: " + filename)
         return retval
@@ -492,7 +499,7 @@ class SbSession:
         :param item: ScienceBase Catalog Item JSON of the Item on which to replace the file
         :return: ScienceBase Catalog Item JSON of the updated Item
         """
-        (path, fname) = os.path.split(filename)
+        fname = os.path.basename(filename)
         #
         # replace file in files list
         #
@@ -651,35 +658,56 @@ class SbSession:
                             retval.append(finfo)
         return retval
 
-    def download_file(self, url, local_filename, destination='.'):
+    def download_file(self, url, local_filename, destination='.', progress_bar=False):
         """Download file from URL
 
         :param url: ScienceBase Catalog Item file download URL
         :param local_filename: Name to use for the local file
         :param destination: Destination directory in which to store the files
+        :param progress_bar: Boolean to turn on progress bar printing
         :return: The full name and path of the downloaded file
         """
         complete_name = os.path.join(destination, local_filename)
         print("downloading " + url + " to " + complete_name)
-        r = self._session.get(url, stream=True)
+        response = self._session.get(url, stream=True)
+
+        # https://stackoverflow.com/a/15645088/3362993
+        download_length = 0
+        if progress_bar is True:
+            try:
+                total_length = int(response.headers.get('content-length'))
+            except Exception:
+                try:
+                    total_length = int(requests.head(url, headers={'Accept-Encoding': None}).headers.get("content-length"))
+                except Exception:
+                    print("No 'content-length' header found to populate progress bar.")
+                    progress_bar=False
 
         with open(complete_name, 'wb') as f:
-            for chunk in r.iter_content(chunk_size=1024):
+            for chunk in response.iter_content(chunk_size=1024):
                 if chunk: # filter out keep-alive new chunks
+                    download_length += len(chunk)
                     f.write(chunk)
                     f.flush()
+                    if progress_bar is True:
+                        done = int(50 * download_length / total_length)
+                        sys.stdout.write("\r[%s%s]" % ('=' * done, ' ' * (50-done)) + " " + str(int(download_length / total_length * 100)) + "%")
+                        sys.stdout.flush()
+            if progress_bar is True:
+                sys.stdout.write('\n')
         return complete_name
 
-    def get_item_files(self, item, destination='.'):
+    def get_item_files(self, item, destination='.', progress_bar=False):
         """Download the individual files attached to a ScienceBase Item
 
         :param item: ScienceBase Catalog Item JSON of the item from which to download files
         :param destination: Destination directory in which to store the files
+        :param progress_bar: Boolean to turn on progress bar printing
         :return: The ScienceBase Catalog file info JSON response
         """
         file_info = self.get_item_file_info(item)
-        for file_info in file_info:
-            self.download_file(file_info['url'], file_info['name'], destination)
+        for finfo in file_info:            
+            self.download_file(finfo['url'], finfo['name'], destination, progress_bar)
         return file_info
 
     def get_my_items_id(self):
@@ -687,15 +715,18 @@ class SbSession:
 
         :return: The ScienceBase Catalog Item ID of the logged in user's My Items folder
         """
-        if (self._username):
+        ret = None
+        if self._username:
             params = {'q': '', 'lq': 'title.untouched:"' + self._username + '"'}
             if self._users_id:
                 params['parentId'] = self._users_id
             items = self.find_items(params)
-            if ('items' in items):
+            if 'items' in items:
                 for item in items['items']:
-                    if (item['title'] == self._username):
-                        return item['id']
+                    if item['title'] == self._username:
+                        ret = item['id']
+                        break
+        return ret
 
     def get_child_ids(self, parentid):
         """Get IDs of all immediate children for a given parent
@@ -934,8 +965,8 @@ class SbSession:
         self._check_errors(response)
         try:
             return response.json()
-        except:
-            raise Exception("Error parsing JSON response: " + response.text)
+        except Exception as exc:
+            raise Exception("Error parsing JSON response: " + response.text) from exc
 
     def _get_text(self, response):
         """Check the status code of the response, and return the text
@@ -946,8 +977,8 @@ class SbSession:
         self._check_errors(response)
         try:
             return response.text
-        except:
-            raise Exception("Error parsing response")
+        except Exception as exc:
+            raise Exception("Error parsing response") from exc
 
     def _check_errors(self, response):
         """Check the status code of the response
@@ -955,16 +986,16 @@ class SbSession:
         :param response: HTTP response to check
         :return: HTTP response, provided an error did not occur in the request
         """
-        if (response.status_code == 404):
+        if response.status_code == 404:
             if "The specified URL cannot be found" in response.text:
                 raise Exception("Request blocked by the USGS web application firewall")
             else:
                 raise Exception("Resource not found, or user does not have access")
-        elif (response.status_code == 401):
+        elif response.status_code == 401:
             raise Exception("Unauthorized access")
-        elif (response.status_code == 429):
+        elif response.status_code == 429:
             raise Exception("Too many requests")
-        elif (response.status_code != 200 and response.status_code != 201):
+        elif response.status_code != 200 and response.status_code != 201:
             if self._retry:
                 return response
             else:
@@ -981,6 +1012,17 @@ class SbSession:
         o = urlparse.urlsplit(url)
         q = [x for x in urlparse.parse_qsl(o.query) if "josso" not in x]
         return urlparse.urlunsplit((o.scheme, o.netloc, o.path, urlencode(q), o.fragment))
+    
+    def _guess_mimetype(self, filename):
+        """Guess mimetype of file
+
+        :param filename: Name of file for which to guess mimetype
+        :return: mimetype of file, or None if it cannot be guessed
+        """
+        mimetype, _ = mimetypes.guess_type(filename)
+        if mimetype is None:
+            mimetype = 'application/octet-stream'
+        return mimetype
 
     def debug(self):
         """Turn on HTTP logging for debugging purposes.
@@ -1098,16 +1140,23 @@ class SbSession:
         """
         return self._update_acls(self.ACL_REMOVE, self.ACL_WRITE, "ROLE:%s" % role_name, item_id)
 
+    def publish_to_public_bucket(self, item_id):
+        """ call publish end point from catalog
+            this should publish all files to public s3 publish bucket
+            TODO: Fix documentation
+        """
+        return self._session.post(self._base_item_url + item_id + "/publishFilesToS3")
+
     def publish_item(self, item_id):
-        """Publish the item, adding PUBLIC read permisisons. User must be USGS or in the publisher role.
-        :param item_id: The ID of the ScienceBase item        
+        """Publish the item, adding PUBLIC read permissions. User must be USGS or in the publisher role.
+        :param item_id: The ID of the ScienceBase item
         :return: The permissions JSON for the given item
         """
         return self._update_acls(self.ACL_ADD, self.ACL_READ, "PUBLIC", item_id)
     
     def unpublish_item(self, item_id):
-        """Unpublish the item, removing PUBLIC read permisisons.
-        :param item_id: The ID of the ScienceBase item        
+        """Unpublish the item, removing PUBLIC read permissions.
+        :param item_id: The ID of the ScienceBase item
         :return: The permissions JSON for the given item
         """
         return self._update_acls(self.ACL_REMOVE, self.ACL_READ, "PUBLIC", item_id)
@@ -1229,15 +1278,15 @@ class SbSession:
         :param reverse: Whether to reverse the relationsip
         :return: ItemLink JSON
         """
-        itemLinkJson = {
+        item_link_json = {
             "itemLinkTypeId": link_type_id
         }
-        itemLinkJson['itemId'] = from_item_id
-        itemLinkJson['relatedItemId'] = to_item_id
+        item_link_json['itemId'] = from_item_id
+        item_link_json['relatedItemId'] = to_item_id
         if reverse:
-            itemLinkJson['reverseRelationship'] = True
+            item_link_json['reverseRelationship'] = True
 
-        ret = self._session.post('%s' % (self._base_item_link_url), data=json.dumps(itemLinkJson))
+        ret = self._session.post(f'{self._base_item_link_url}', data=json.dumps(item_link_json))
         return self._get_json(ret)
 
     def create_related_item_link(self, from_item_id, to_item_id):
@@ -1249,4 +1298,3 @@ class SbSession:
         """
         related_item_link = self.get_item_link_type_by_name('related')
         return self.create_item_link(from_item_id, to_item_id, related_item_link['id'])
-    
